/* **********************************************************************
*Copyright (C) 1999 Elliott Mitchell					*
*Elliott Mitchell takes no responsibility for failures, or damage done	*
*by this program.							*
*									*
*$Id: pascal.c,v 1.37 2003/09/17 07:33:21 ehem Exp $			*
************************************************************************/

#include <stdio.h>
#include <X11/Xlib.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <GL/gl.h>
#include <GL/glx.h>
#include <GL/glu.h>
#include <GL/glut.h>
#include "pascal.h"

#ifndef SEED
#define SEED 0x36FA380B
#endif

#define FONT GLUT_BITMAP_HELVETICA_12
#ifdef FONTTEST
void initfont();
#endif

#ifdef NOFOG
#undef USE_FOG
#else
#ifndef USE_FOG
#define USE_FOG
#endif
#endif


#ifndef LOCX
#define LOCX 0.5
#endif
#ifndef LOCY
#define LOCY 0.5
#endif
GLfloat loc[2]={LOCX, LOCY};

#define MAXOBJ 9

#ifndef DEFAULTDEPTH
#define DEFAULTDEPTH 4
#endif
/*Maximum depth we'll go to*/
GLuint maxdepth;
/*Current base display list*/
GLuint curstart, curskip;

GLuint fore;

GLuint curdepth=DEFAULTDEPTH;

GLdouble Wangle=1.0, Xangle=0.0, Yangle=0.0, Zangle=0.0;
int lastX, lastY;
int height, width;
GLfloat lightZeroPosition[] = {10.0, 4.0, 10.0, 1.0};
GLfloat lightZeroColor[] = {0.8, 1.0, 0.8, 1.0}; /*greenish*/
GLfloat lightOnePosition[] = {-1.0, -2.0, 1.0, 0.0};
GLfloat lightOneColor[] = {0.6, 0.3, 0.2, 1.0}; /*redish*/


struct object {
	void (*init)(struct object *);
	GLuint start, skip;
	GLuint low, high;
};


struct point {
	GLfloat n[3];
	GLfloat v[3];
};

struct vpoint {
	GLfloat v[3];
};

struct polyhedra {
	void (*init)(struct polyhedra *);
	GLuint start, skip;
	GLuint low, high;

	unsigned center:1; /*needs the center or not?*/
	void (*init2)(struct vpoint *, GLuint *);

	GLuint pcnt, fcnt, fsz;

	struct vpoint *pts;
	GLuint *face;
};

void _init4(struct vpoint *, GLuint *),
_init6(struct vpoint *, GLuint *),
_init8(struct vpoint *, GLuint *),
_init12(struct vpoint *, GLuint *),
_init20(struct vpoint *, GLuint *);

void makepolydls(struct polyhedra *);

struct polyhedra regobjs[5]={
{makepolydls,	0, 1, 0, 0,	0,  _init4,  4,  4, 3, NULL, NULL},
{makepolydls,	0, 1, 0, 0,	1,  _init6,  8,  6, 4, NULL, NULL},
{makepolydls,	0, 1, 0, 0,	1,  _init8,  6,  8, 3, NULL, NULL},
{makepolydls,	0, 1, 0, 0,	1, _init12, 20, 12, 5, NULL, NULL},
{makepolydls,	0, 1, 0, 0,	1, _init20, 12, 20, 3, NULL, NULL},
};

void mk1_25(struct object *);
struct object _obj1_25={mk1_25, 0, 0, 0, 0};

struct tessobj {
	struct object obj;
	struct polyhedra *poly;
};

void mktessobj(struct tessobj *);
struct tessobj tessobjs[3]={
{{(void (*)(struct object *))mktessobj,	0, 1, 0, 0},	regobjs+0},
{{(void (*)(struct object *))mktessobj,	0, 1, 0, 0},	regobjs+2},
{{(void (*)(struct object *))mktessobj,	0, 1, 0, 0},	regobjs+4},
};

struct object *objs[MAXOBJ+1]={(struct object *)(regobjs+0),
(struct object *)(regobjs+1), (struct object *)(regobjs+2),
(struct object *)(regobjs+3), (struct object *)(regobjs+4), &_obj1_25,
&tessobjs[0].obj, &tessobjs[1].obj, &tessobjs[2].obj, NULL};


int startup(int, char **);

void showmatr(GLenum);
void regen(void);

void redisplay(void);
#define POSTFULLREDISPLAY glutIdleFunc(redisplay)

void draw(void), hresize(int, int), hkeyb1(unsigned char, int, int),
hkeyb2(int, int, int), hmouseb(int, int, int, int), hmousem(int, int);

void init(void);
int objinit(void);


/* **********************************************************************
*Main Entry Point							*
************************************************************************/

int main(int argc, char **argv)
{
	srandom(SEED);

	{
		register int i;

		if((i=startup(argc, argv))) return(i);

		for(i=0; i<MAXOBJ; ++i) objs[i]->init(objs[i]);
	}

#ifdef FONTTEST
	initfont();
#endif

	curstart=objs[0]->start;
	curskip=objs[0]->skip;
	regen();

	glutMainLoop();
	return(0);
}

/* **********************************************************************
*Regular Object Initialization						*
************************************************************************/

void makepolydls(struct polyhedra *this)
{
	GLuint a, b, c;
	GLuint *facep;

	struct vpoint *pts;
	GLuint *face;

	/*Given that stacks get grown dynamically, I expect it to either
successfully grow, or for the OS memory allocation to fail and us to get
terminated**
	pts=alloca(sizeof(struct vpoint)*this->pcnt);
	face=alloca(sizeof(GLuint)*this->fcnt*this->fsz);*/

	this->pts=pts=malloc(sizeof(struct vpoint)*this->pcnt);
	this->face=face=malloc(sizeof(GLuint)*this->fcnt*this->fsz);

	if(!pts||!face) {
		fprintf(stderr, "Out of memory\n");
		exit(1);
	}

	this->init2(pts, face);


	a=glGenLists(maxdepth);
	this->low=a;
	this->high=a+maxdepth-1;
	this->start=a;
/*	this->skip=1; (already initialized)*/
	c=a+maxdepth;

#ifndef SMOOTH
	glVertexPointer(3, GL_FLOAT, 0, (void *)pts);
#else
	glVertexPointer(3, GL_FLOAT, 0, (void *)pts);
	glNormalPointer(GL_FLOAT, 0, (void *)pts);
#endif

	facep=face;

	/*Setup the simple polyhedra display list*/
	glNewList(a, GL_COMPILE);
#ifndef SMOOTH
	glDisableClientState(GL_NORMAL_ARRAY);
#endif
	for(b=0; b<this->fcnt; ++b) {
#ifndef SMOOTH
		GLdouble norm[3]={0.0, 0.0, 0.0};
		register long i;
		for(i=0; i<this->fsz; ++i) {
			norm[0]+=pts[facep[i]].v[0];
			norm[1]+=pts[facep[i]].v[1];
			norm[2]+=pts[facep[i]].v[2];
		}
		norm[0]/=this->fsz;
		norm[1]/=this->fsz;
		norm[2]/=this->fsz;

		glNormal3dv(norm);
#endif

		glDrawElements(GL_POLYGON, this->fsz, GL_UNSIGNED_INT, facep);
		facep+=this->fsz;
	}
#ifndef SMOOTH
	glEnableClientState(GL_NORMAL_ARRAY);
#endif
	glEndList();


	if(this->center) {
		for(++a; a<c; ++a) {
			glNewList(a, GL_COMPILE);

			glPushMatrix();
			glScalef(1.0/3.0, 1.0/3.0, 1.0/3.0);
			glCallList(a-1);
			glPopMatrix();

			for(b=0; b<this->pcnt; ++b) {
				glPushMatrix();

				glScalef(1.0/3.0, 1.0/3.0, 1.0/3.0);
				glTranslatef(2.0*pts[b].v[0],
2.0*pts[b].v[1], 2.0*pts[b].v[2]);

				glCallList(a-1);
				glPopMatrix();
			}

			glEndList();
		}
	} else {
		for(++a; a<c; ++a) {
			glNewList(a, GL_COMPILE);

			for(b=0; b<this->pcnt; ++b) {
				glPushMatrix();

				glScalef(0.5, 0.5, 0.5);
				glTranslatef(pts[b].v[0],
pts[b].v[1], pts[b].v[2]);

				glCallList(a-1);
				glPopMatrix();
			}

			glEndList();
		}
	}
}

void _init4(struct vpoint *pts, GLuint *face)
{
	int a, b;

/*nasty values needed to create the tetrahedra*/
	double elen, magicv;

	const GLuint faces[12]={0, 1, 3,   0, 3, 2,   0, 2, 1,   1, 2, 3};

	for(a=0; a<12; ++a) face[a]=faces[a];

	/*everybody's favorite: the reciprocal of the square root of two*/
	magicv=sqrt(2)*0.5;

	/*and everybody's favorite 3D Pythagorean*/
	elen=1.0/sqrt(1.0*1.0+magicv*magicv);

#define AX0 0
#define AX1 1
#define AX2 2

	pts[1].v[AX0]=-elen; pts[0].v[AX0]=elen;
	pts[1].v[AX1]=pts[0].v[AX1]=0.0;
	pts[3].v[AX0]=pts[2].v[AX0]=0.0;
	pts[3].v[AX1]=-elen; pts[2].v[AX1]=elen;

	pts[1].v[AX2]=pts[0].v[AX2]=magicv*elen;
	pts[3].v[AX2]=pts[2].v[AX2]=-magicv*elen;
}

void _init6(struct vpoint *pts, GLuint *face)
{
	int a, b;

	double magicv;

	const GLuint faces[24]={0, 1, 3, 2,	0, 2, 6, 4,	0, 4, 5, 1,
1, 5, 7, 3,	2, 3, 7, 6,	4, 6, 7, 5};

	for(a=0; a<24; ++a) face[a]=faces[a];

	/*and everybody's favorite 3D Pythagorean, (to make it a unit cube*/
	magicv=1.0/sqrt(1.0+1.0+1.0);

	pts[0].v[2]= (pts[0].v[1]= (pts[0].v[0]= magicv));
	pts[1].v[2]= (pts[1].v[1]=-(pts[1].v[0]=-magicv));
	pts[2].v[2]=-(pts[2].v[1]=-(pts[2].v[0]= magicv));
	pts[3].v[2]=-(pts[3].v[1]= (pts[3].v[0]=-magicv));
	pts[4].v[2]=-(pts[4].v[1]= (pts[4].v[0]= magicv));
	pts[5].v[2]=-(pts[5].v[1]=-(pts[5].v[0]=-magicv));
	pts[6].v[2]= (pts[6].v[1]=-(pts[6].v[0]= magicv));
	pts[7].v[2]= (pts[7].v[1]= (pts[7].v[0]=-magicv));

#undef AX0
#undef AX1
#undef AX2
}

void _init8(struct vpoint *pts, GLuint *face)
{
	int a, b;

	const GLuint faces[24]={0, 1, 2,   0, 2, 3,   0, 3, 4,   0, 4, 1,
1, 5, 2,   2, 5, 3,   3, 5, 4,   4, 5, 1};

	for(a=0; a<24; ++a) face[a]=faces[a];

	pts[0].v[0]=1.0;
	pts[0].v[2]=pts[0].v[1]=0.0;

	pts[1].v[1]=1.0;
	pts[1].v[2]=pts[1].v[0]=0.0;

	pts[2].v[2]=1.0;
	pts[2].v[1]=pts[2].v[0]=0.0;

	pts[3].v[1]=-1.0;
	pts[3].v[2]=pts[3].v[0]=0.0;

	pts[4].v[2]=-1.0;
	pts[4].v[1]=pts[4].v[0]=0.0;

	pts[5].v[0]=-1.0;
	pts[5].v[2]=pts[5].v[1]=0.0;
}

void _init12(struct vpoint *pts, GLuint *face)
{
	int a, b;

	/*ang1=54  ang2=36*/
	const GLdouble ang1=M_PI*3.0/10.0, ang2=M_PI*1.0/5.0;
	GLdouble wcn, wcf, hcn, hcf; /*more helper constants*/

	GLdouble S, L, h, H, hT, HT;
	GLdouble RADpmn, RADpmx, RADpdiff, RADtmn, RADtmx, RADtdiff;

	const GLuint faces[60]={ 0,  1,  2,  3,  4,	 0,  5, 10,  6,  1,
 1,  6, 11,  7,  2,	 2,  7, 12,  8,  3,	 3,  8, 13,  9,  4,
 4,  9, 14,  5,  0,	 5, 14, 19, 15, 10,	 6, 10, 15, 16, 11,
 7, 11, 16, 17, 12,	 8, 12, 17, 18, 13,	 9, 13, 18, 19, 14,
15, 19, 18, 17, 16};

	for(a=0; a<60; ++a) face[a]=faces[a];

	S=1.0;

	L=2*S*sin(ang1);
	RADpmx=0.5*L/sin(ang2);
	RADpmn=0.5*L/tan(ang2);
	RADpdiff=RADpmx-RADpmn;
	h=S*cos(ang1);
	H=sqrt(h*h-RADpdiff*RADpdiff)*0.5;

	RADtmx=0.5*S/sin(ang2);
	RADtmn=0.5*S/tan(ang2);
	RADtdiff=RADtmx-RADtmn;
	hT=S;
	/*since we can use it, 'S' is used as a temporary*/
	HT=sqrt(hT*hT-(S=RADpmx-RADtmx)*S)+H;


	/*S is now recycled as a scaling factor*/
	/*Everybody remember their Pythagorean?*/
	S=1.0/sqrt(H*H+RADpmx*RADpmx);

	/*Thankfully we don't have to recompute the above formulae*/
	RADpmx*=S;
	H*=S;
	RADtmx*=S;
	HT*=S;

	wcn=sin(ang2*2);
	wcf=sin(ang2*4);
	hcn=cos(ang2*2);
	hcf=cos(ang2*4);


	pts[ 0].v[0]=RADtmx;
	pts[ 0].v[1]=0.0;
	pts[ 0].v[2]=HT;

	pts[ 1].v[0]=RADtmx*hcn;
	pts[ 1].v[1]=RADtmx*wcn;
	pts[ 1].v[2]=HT;

	pts[ 2].v[0]=RADtmx*hcf;
	pts[ 2].v[1]=RADtmx*wcf;
	pts[ 2].v[2]=HT;

	pts[ 3].v[0]=RADtmx*hcf;
	pts[ 3].v[1]=RADtmx*-wcf;
	pts[ 3].v[2]=HT;

	pts[ 4].v[0]=RADtmx*hcn;
	pts[ 4].v[1]=RADtmx*-wcn;
	pts[ 4].v[2]=HT;


	pts[ 5].v[0]=RADpmx;
	pts[ 5].v[1]=0.0;
	pts[ 5].v[2]=H;

	pts[ 6].v[0]=RADpmx*hcn;
	pts[ 6].v[1]=RADpmx*wcn;
	pts[ 6].v[2]=H;

	pts[ 7].v[0]=RADpmx*hcf;
	pts[ 7].v[1]=RADpmx*wcf;
	pts[ 7].v[2]=H;

	pts[ 8].v[0]=RADpmx*hcf;
	pts[ 8].v[1]=RADpmx*-wcf;
	pts[ 8].v[2]=H;

	pts[ 9].v[0]=RADpmx*hcn;
	pts[ 9].v[1]=RADpmx*-wcn;
	pts[ 9].v[2]=H;



	pts[10].v[0]=RADpmx*-hcf;
	pts[10].v[1]=RADpmx*wcf;
	pts[10].v[2]=-H;

	pts[11].v[0]=RADpmx*-hcn;
	pts[11].v[1]=RADpmx*wcn;
	pts[11].v[2]=-H;

	pts[12].v[0]=-RADpmx;
	pts[12].v[1]=0.0;
	pts[12].v[2]=-H;

	pts[13].v[0]=RADpmx*-hcn;
	pts[13].v[1]=RADpmx*-wcn;
	pts[13].v[2]=-H;

	pts[14].v[0]=RADpmx*-hcf;
	pts[14].v[1]=RADpmx*-wcf;
	pts[14].v[2]=-H;


	pts[15].v[0]=RADtmx*-hcf;
	pts[15].v[1]=RADtmx*wcf;
	pts[15].v[2]=-HT;

	pts[16].v[0]=RADtmx*-hcn;
	pts[16].v[1]=RADtmx*wcn;
	pts[16].v[2]=-HT;

	pts[17].v[0]=-RADtmx;
	pts[17].v[1]=0.0;
	pts[17].v[2]=-HT;

	pts[18].v[0]=RADtmx*-hcn;
	pts[18].v[1]=RADtmx*-wcn;
	pts[18].v[2]=-HT;

	pts[19].v[0]=RADtmx*-hcf;
	pts[19].v[1]=RADtmx*-wcf;
	pts[19].v[2]=-HT;
}

void _init20(struct vpoint *pts, GLuint *face)
{
	int a, b;

	/*ang1=60  ang2=36*/
	const GLdouble ang1=M_PI*1.0/3.0, ang2=M_PI*1.0/5.0;

	GLdouble S, h, H, RAD, RADpmn, RADpmx, RADpdiff;

	const GLuint faces[60]={ 0,  1,  2,	 0,  2,  3,	 0,  3,  4,
 0,  4,  5,	 0,  5,  1,	 1,  6,  2,	 2,  7,  3,	 3,  8,  4,
 4,  9,  5,	 5, 10,  1,	 1, 10,  6,	 2,  6,  7,	 3,  7,  8,
 4,  8,  9,	 5,  9, 10,	 6, 10, 11,	 7,  6, 11,	 8,  7, 11,
 9,  8, 11,	10,  9, 11};

	for(a=0; a<60; ++a) face[a]=faces[a];

	S=1.0;

	RADpmn=0.5*S/tan(ang2);
	RADpmx=0.5*S/sin(ang2);
	RADpdiff=RADpmx-RADpmn;

	h=S*sin(ang1);
	H=sqrt(h*h-RADpdiff*RADpdiff)*0.5;

	RAD=sqrt(RADpmx*RADpmx+H*H);


	/*S is now recycled as a scaling factor*/
	/*Everybody remember their Pythagorean?*/
	S=1.0/RAD;

	/*Thankfully we don't have to recompute the above formulae*/
	RADpmx*=S;
	H*=S;


	pts[ 0].v[0]=0.0;
	pts[ 0].v[1]=0.0;
	pts[ 0].v[2]=RAD;


	pts[ 1].v[0]=RADpmx;
	pts[ 1].v[1]=0.0;
	pts[ 1].v[2]=H;

	pts[ 2].v[0]=RADpmx*cos(ang2*2.0);
	pts[ 2].v[1]=RADpmx*sin(ang2*2.0);
	pts[ 2].v[2]=H;

	pts[ 3].v[0]=RADpmx*cos(ang2*4.0);
	pts[ 3].v[1]=RADpmx*sin(ang2*4.0);
	pts[ 3].v[2]=H;

	pts[ 4].v[0]=RADpmx*cos(ang2*6.0);
	pts[ 4].v[1]=RADpmx*sin(ang2*6.0);
	pts[ 4].v[2]=H;

	pts[ 5].v[0]=RADpmx*cos(ang2*8.0);
	pts[ 5].v[1]=RADpmx*sin(ang2*8.0);
	pts[ 5].v[2]=H;



	pts[ 6].v[0]=RADpmx*cos(ang2*1.0);
	pts[ 6].v[1]=RADpmx*sin(ang2*1.0);
	pts[ 6].v[2]=-H;

	pts[ 7].v[0]=RADpmx*cos(ang2*3.0);
	pts[ 7].v[1]=RADpmx*sin(ang2*3.0);
	pts[ 7].v[2]=-H;

	pts[ 8].v[0]=-RADpmx;
	pts[ 8].v[1]=0.0;
	pts[ 8].v[2]=-H;

	pts[ 9].v[0]=RADpmx*cos(ang2*7.0);
	pts[ 9].v[1]=RADpmx*sin(ang2*7.0);
	pts[ 9].v[2]=-H;

	pts[10].v[0]=RADpmx*cos(ang2*9.0);
	pts[10].v[1]=RADpmx*sin(ang2*9.0);
	pts[10].v[2]=-H;


	pts[11].v[0]=0.0;
	pts[11].v[1]=0.0;
	pts[11].v[2]=-RAD;
}

/* **********************************************************************
*Tesselated Object Initialization					*
************************************************************************/

static void _tesstri(int, struct vpoint *, struct vpoint *, struct vpoint *);
void mktessobj(struct tessobj *this)
{
#define TESSMAX 6
	GLuint lists, i, j;

	if(this->poly->fsz!=3) {
		fprintf(stderr, "Internal error #%08d\n", __LINE__);
		exit(1);
	}

	lists=glGenLists(maxdepth);

	this->obj.low=lists;
	this->obj.high=lists+maxdepth-1;
	this->obj.start=lists;
/*	this->obj.skip=1; (already initialized)*/

	for(i=0; i<TESSMAX; ++i) {
		glNewList(lists+i, GL_COMPILE);
		glBegin(GL_TRIANGLES);

		for(j=0; j<this->poly->fcnt; ++j) _tesstri(i,
this->poly->pts+this->poly->face[j*this->poly->fsz+0],
this->poly->pts+this->poly->face[j*this->poly->fsz+1],
this->poly->pts+this->poly->face[j*this->poly->fsz+2]);

		glEnd();
		glEndList();
	}

	for(i=TESSMAX,j=0; i<maxdepth; ++i) {
		if(j>=TESSMAX) j-=TESSMAX;
		glNewList(lists+i, GL_COMPILE);
		glCallList(lists+j);
		glEndList();
	}
}

void _tesstri(int level, struct vpoint *a, struct vpoint *b, struct vpoint *c)
{
	if(level) { /*tesselate*/
		double mul;
		struct vpoint a0, b0, c0;

		mul=0.5;

		a0.v[0]=(a->v[0]+b->v[0])*mul;
		a0.v[1]=(a->v[1]+b->v[1])*mul;
		a0.v[2]=(a->v[2]+b->v[2])*mul;

		b0.v[0]=(c->v[0]+b->v[0])*mul;
		b0.v[1]=(c->v[1]+b->v[1])*mul;
		b0.v[2]=(c->v[2]+b->v[2])*mul;

		c0.v[0]=(a->v[0]+c->v[0])*mul;
		c0.v[1]=(a->v[1]+c->v[1])*mul;
		c0.v[2]=(a->v[2]+c->v[2])*mul;

		/*Note, removing the "1.0/" gives an interesting effect*/
		mul=1.0/sqrt(a0.v[0]*a0.v[0]+a0.v[1]*a0.v[1]+a0.v[2]*a0.v[2]);
		a0.v[0]*=mul;  a0.v[1]*=mul;  a0.v[2]*=mul;
		mul=1.0/sqrt(b0.v[0]*b0.v[0]+b0.v[1]*b0.v[1]+b0.v[2]*b0.v[2]);
		b0.v[0]*=mul;  b0.v[1]*=mul;  b0.v[2]*=mul;
		mul=1.0/sqrt(c0.v[0]*c0.v[0]+c0.v[1]*c0.v[1]+c0.v[2]*c0.v[2]);
		c0.v[0]*=mul;	c0.v[1]*=mul;	c0.v[2]*=mul;

		--level;
		_tesstri(level, a, &a0, &c0);
		_tesstri(level, b, &b0, &a0);
		_tesstri(level, c, &c0, &b0);
		_tesstri(level, &a0, &b0, &c0);
	} else { /*simply draw the side*/
/*		glBegin(GL_TRIANGLES);  (done in parent)*/
		glNormal3fv(a->v);
		glVertex3fv(a->v);
		glNormal3fv(b->v);
		glVertex3fv(b->v);
		glNormal3fv(c->v);
		glVertex3fv(c->v);
/*		glEnd();*/
	}
}

/* **********************************************************************
*Non-Regular Object Initialization					*
************************************************************************/

void mk1_25(struct object *this)
{
#define AX0 0
#define AX1 1
#define AX2 2

/*Turn Left*/
#define AXT0 0.0
#define AXT1 1.0
#define AXT2 0.0

/*Pitch Up*/
#define AXP0 1.0
#define AXP1 0.0
#define AXP2 0.0

/*Roll Left*/
#define AXR0 0.0
#define AXR1 0.0
#define AXR2 1.0

/* **********************************************************************
*The L-system productions in the notation of the book:			*
*									*
*	Angle:	22.5 (deg)						*
*	Start:	A							*
*	p1:	A->[&FL!A]/////'[&FL!A]///////'[&FL!A]			*
*	p2:	F->S/////F						*
*	p3:	S->F L							*
*	p4:	L->['''^^{-f+f+f-|-f+f+f}]				*
*									*
*Terminal productions: F, f						*
*Nonrecursive productions: L						*
*Nonterminal productions: A, F, S					*
************************************************************************/
#define ANG 22.5
#define UNIT 1.0
#define UNITX 0.0
#define UNITY 0.0
#define UNITZ 1.0
#define STEM 0.1

#define SKIP 4
#define SCALEINCR scaleincr

	const GLfloat green[3]={0.000, 0.392, 0.000},
brown[3]={0.545, 0.271, 0.075};

	GLuint low;
	GLuint prodL, prodF;
	GLuint baseA, baseS, baseF;
	GLuint tmp, tmp2;

	static struct point pts[16];

	GLuint baseCall;
	GLfloat scale=1.0;

	GLfloat scaleincr;
	scaleincr=0.5/cos(ANG*(M_PI/180.0));


	low=glGenLists(2+(maxdepth-2)*SKIP);

	this->low=low;
	this->high=2+(maxdepth-2)*SKIP;

	prodF=low+0;
	prodL=low+1;

	baseF=low+2;
	baseS=low+3;
	baseA=low+4;
	baseCall=low+5;

	this->start=baseCall;
	this->skip=SKIP;

	/* STEM */

	{ /*point generation*/
		GLdouble hf;

		hf=sqrt(2)*0.5*STEM;

		pts[ 0].n[AX0]=pts[ 0].v[AX0]=STEM;
		pts[ 0].n[AX1]=pts[ 0].v[AX1]=0.0;
		pts[ 1].n[AX0]=pts[ 1].v[AX0]=STEM;
		pts[ 1].n[AX1]=pts[ 1].v[AX1]=0.0;

		pts[ 2].n[AX0]=pts[ 2].v[AX0]=hf;
		pts[ 2].n[AX1]=pts[ 2].v[AX1]=hf;
		pts[ 3].n[AX0]=pts[ 3].v[AX0]=hf;
		pts[ 3].n[AX1]=pts[ 3].v[AX1]=hf;

		pts[ 4].n[AX0]=pts[ 4].v[AX0]=0.0;
		pts[ 4].n[AX1]=pts[ 4].v[AX1]=STEM;
		pts[ 5].n[AX0]=pts[ 5].v[AX0]=0.0;
		pts[ 5].n[AX1]=pts[ 5].v[AX1]=STEM;

		pts[ 6].n[AX0]=pts[ 6].v[AX0]=-hf;
		pts[ 6].n[AX1]=pts[ 6].v[AX1]=hf;
		pts[ 7].n[AX0]=pts[ 7].v[AX0]=-hf;
		pts[ 7].n[AX1]=pts[ 7].v[AX1]=hf;


		pts[ 8].n[AX0]=pts[ 8].v[AX0]=-STEM;
		pts[ 8].n[AX1]=pts[ 8].v[AX1]=0.0;
		pts[ 9].n[AX0]=pts[ 9].v[AX0]=-STEM;
		pts[ 9].n[AX1]=pts[ 9].v[AX1]=0.0;

		pts[10].n[AX0]=pts[10].v[AX0]=-hf;
		pts[10].n[AX1]=pts[10].v[AX1]=-hf;
		pts[11].n[AX0]=pts[11].v[AX0]=-hf;
		pts[11].n[AX1]=pts[11].v[AX1]=-hf;

		pts[12].n[AX0]=pts[12].v[AX0]=0.0;
		pts[12].n[AX1]=pts[12].v[AX1]=-STEM;
		pts[13].n[AX0]=pts[13].v[AX0]=0.0;
		pts[13].n[AX1]=pts[13].v[AX1]=-STEM;

		pts[14].n[AX0]=pts[14].v[AX0]=hf;
		pts[14].n[AX1]=pts[14].v[AX1]=-hf;
		pts[15].n[AX0]=pts[15].v[AX0]=hf;
		pts[15].n[AX1]=pts[15].v[AX1]=-hf;

		pts[ 0].n[AX2]=pts[ 0].v[AX2]=0.0;
		pts[ 2].n[AX2]=pts[ 2].v[AX2]=0.0;
		pts[ 4].n[AX2]=pts[ 4].v[AX2]=0.0;
		pts[ 6].n[AX2]=pts[ 6].v[AX2]=0.0;
		pts[ 8].n[AX2]=pts[ 8].v[AX2]=0.0;
		pts[10].n[AX2]=pts[10].v[AX2]=0.0;
		pts[12].n[AX2]=pts[12].v[AX2]=0.0;
		pts[14].n[AX2]=pts[14].v[AX2]=0.0;

		pts[ 1].n[AX2]=pts[ 1].v[AX2]=UNIT;
		pts[ 3].n[AX2]=pts[ 3].v[AX2]=UNIT;
		pts[ 5].n[AX2]=pts[ 5].v[AX2]=UNIT;
		pts[ 7].n[AX2]=pts[ 7].v[AX2]=UNIT;
		pts[ 9].n[AX2]=pts[ 9].v[AX2]=UNIT;
		pts[11].n[AX2]=pts[11].v[AX2]=UNIT;
		pts[13].n[AX2]=pts[13].v[AX2]=UNIT;
		pts[15].n[AX2]=pts[15].v[AX2]=UNIT;
	}

	glInterleavedArrays(GL_N3F_V3F, 0, (void *)pts);

	glNewList(prodF, GL_COMPILE);
	glColor3fv(brown);
	for(tmp=0, tmp2=14; tmp<15; tmp2=tmp, tmp+=2) {
		GLuint face[4]={tmp, tmp+1, tmp2+1, tmp2};
		glDrawElements(GL_POLYGON, 4, GL_UNSIGNED_INT, face);
	}
	glTranslatef(UNITX, UNITY, UNITZ);
	glEndList();

	/*Leaves (production 3)*/
	glNewList(prodL, GL_COMPILE);
	glColor3fv(green);
	glPushAttrib(GL_POLYGON_BIT);
	glDisable(GL_CULL_FACE);

	glPushMatrix();
/*L->['''^^{-f+f+f-|-f+f+f}], note we compress multiple rotations into one*/
	glRotatef(ANG*2.0, AXP0, AXP1, AXP2);
	{
#define LAX0 AX0
#define LAX1 AX2
		const double angrad=ANG*M_PI/180.0;
		GLfloat loc[3]={0.0, 0.0, 0.0};
		double dir=0.0;

		glBegin(GL_POLYGON);
		/*Note this is a little bit of cruft here...*/
		glNormal3f(AXT0, AXT1, AXT2);

		glVertex3fv(loc);

		dir-=angrad;
		loc[LAX0]+=UNIT*cos(dir);
		loc[LAX1]+=UNIT*sin(dir);

		glVertex3fv(loc);

		dir+=angrad;
		loc[LAX0]+=UNIT*cos(dir);
		loc[LAX1]+=UNIT*sin(dir);

		glVertex3fv(loc);

		dir+=angrad;
		loc[LAX0]+=UNIT*cos(dir);
		loc[LAX1]+=UNIT*sin(dir);

		glVertex3fv(loc);

		dir-=angrad*2.0+M_PI;
		loc[LAX0]+=UNIT*cos(dir);
		loc[LAX1]+=UNIT*sin(dir);

		glVertex3fv(loc);

		dir+=angrad;
		loc[LAX0]+=UNIT*cos(dir);
		loc[LAX1]+=UNIT*sin(dir);

		glVertex3fv(loc);

		dir+=angrad;
		loc[LAX0]+=UNIT*cos(dir);
		loc[LAX1]+=UNIT*sin(dir);

		glVertex3fv(loc);

		glEnd();
#undef LAX0
#undef LAX1
	}
	glPopMatrix();

	glPopAttrib();
	glEndList();

	/*************************"Boot" Display Lists******************/

	glNewList(baseF, GL_COMPILE);
	glRotatef(ANG*5.0, -AXR0, -AXR1, -AXR2);
	glCallList(prodF);
	glEndList();

	glNewList(baseS, GL_COMPILE);
	glCallList(prodF);
	glCallList(prodL);
	glEndList();


	glNewList(baseA, GL_COMPILE);
	glPushMatrix();
	glRotatef(ANG, -AXP0, -AXP1, -AXP2);
	glCallList(prodF);
	glCallList(prodL);
	glPopMatrix();

	glRotatef(ANG*5.0, -AXR0, -AXR1, -AXR2);
	glPushMatrix();
	glRotatef(ANG, -AXP0, -AXP1, -AXP2);
	glCallList(prodF);
	glCallList(prodL);
	glPopMatrix();

	glRotatef(ANG*7.0, -AXR0, -AXR1, -AXR2);
	glPushMatrix();
	glRotatef(ANG, -AXP0, -AXP1, -AXP2);
	glCallList(prodF);
	glCallList(prodL);
	glPopMatrix();
	glEndList();


	glNewList(baseCall, GL_COMPILE);
	glScalef(scale, scale, scale);
	glCallList(baseA);
	glEndList();


	for(tmp=0; tmp<maxdepth-SKIP; ++tmp) {

		glNewList(baseF+SKIP, GL_COMPILE);
		glCallList(baseS);
		glRotatef(ANG*5.0, -AXR0, -AXR1, -AXR2);
		glCallList(baseF);
		glEndList();

		glNewList(baseS+SKIP, GL_COMPILE);
		glCallList(baseF);
		glCallList(prodL);
		glEndList();


		glNewList(baseA+SKIP, GL_COMPILE);
		glPushMatrix();
		glRotatef(ANG, -AXP0, -AXP1, -AXP2);
		glCallList(baseF);
		glCallList(prodL);
		glCallList(baseA);
		glPopMatrix();

		glRotatef(ANG*5.0, -AXR0, -AXR1, -AXR2);
		glPushMatrix();
		glRotatef(ANG, -AXP0, -AXP1, -AXP2);
		glCallList(baseF);
		glCallList(prodL);
		glCallList(baseA);
		glPopMatrix();

		glRotatef(ANG*7.0, -AXR0, -AXR1, -AXR2);
		glPushMatrix();
		glRotatef(ANG, -AXP0, -AXP1, -AXP2);
		glCallList(baseF);
		glCallList(prodL);
		glCallList(baseA);
		glPopMatrix();
		glEndList();


		baseF+=SKIP;
		baseA+=SKIP;
		baseS+=SKIP;


		glNewList(baseCall+=SKIP, GL_COMPILE);
		scale*=SCALEINCR;
		glScalef(scale, scale, scale);
		glCallList(baseA);
		glEndList();
	}

#undef AX0
#undef AX1
#undef AX2

#undef AXT0
#undef AXT1
#undef AXT2

#undef AXP0
#undef AXP1
#undef AXP2

#undef AXR0
#undef AXR1
#undef AXR2

#undef ANG
#undef UNIT
#undef UNITX
#undef UNITY
#undef UNITZ
#undef STEM

#undef SKIP
#undef SCALEINCR
}


/* **********************************************************************
*Runtime Event Functions						*
************************************************************************/

void hkeyb1(unsigned char key, int x, int y)
{
	switch(key) {
	case 0x1B: /*escape key*/
		exit(0);
	break;
	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
		{
			GLuint new;

			new=key-'0';
			if(new!=curdepth) {
				curdepth=new;
				glutPostRedisplay();
			}
		}
	break;
	default:
		fprintf(stderr, "Unknown normal keycode: 0x%X\n", key);
	break;
	}
}

void hkeyb2(int key, int x, int y)
{
	switch(key) {
	case GLUT_KEY_F1:
	case GLUT_KEY_F2:
	case GLUT_KEY_F3:
	case GLUT_KEY_F4:
	case GLUT_KEY_F5:
	case GLUT_KEY_F6:
	case GLUT_KEY_F7:
	case GLUT_KEY_F8:
	case GLUT_KEY_F9:
	case GLUT_KEY_F10:
		key-=GLUT_KEY_F1;

		if(key>(MAXOBJ-1)) {
			fprintf(stderr, "Unknown Object %d\n", key+1);
			break;
		}

		curstart=objs[key]->start;
		curskip=objs[key]->skip;
		glutPostRedisplay();
	break;


	case GLUT_KEY_F11:
		fprintf(stderr, "\n");
		showmatr(GL_PROJECTION_MATRIX);
		showmatr(GL_MODELVIEW_MATRIX);
	break;
	case GLUT_KEY_F12:
		fprintf(stderr, "Rotation: [%+f, (%+f, %+f, %+f)]\n", Wangle,
Xangle, Yangle, Zangle);
	break;
	default:
		fprintf(stderr, "Unknown special keycode: 0x%X\n", key);
	break;
	}
}

void hmouseb(int button, int st, int x, int y)
{
	if(button!=GLUT_LEFT_BUTTON) return;

	lastX=x;
	lastY=y;
}

#ifndef USEABLE_ANGLE
/*nasty function to take care of nastiness below*/
static GLdouble fixup_angle(GLdouble ang, GLdouble old, GLdouble sin,
GLdouble cos)
{
	if((sin>=0)^(cos>=0)) {
		if((old>=M_PI*0.25)&&(old<M_PI*1.25)) {
			if(ang<M_PI*0.25) ang+=M_PI;
		} else {
			if(ang>=M_PI*0.25) ang+=M_PI;
		}
	} else {
		if((old>=M_PI*0.75)&&(old<M_PI*1.75)) {
			if(ang<M_PI*0.75) ang+=M_PI;
		} else {
			if(ang>=M_PI*0.75) ang+=M_PI;
		}
	}
	return(ang);
}
#endif

void hmousem(int x, int y)
{
	GLdouble dx, dy;
#ifndef USE_QUATERNIONS
#ifndef USEABLE_ANGLE
	GLdouble tmpX, tmpY, tmpZ;
#endif
#else
	GLdouble tmpW, tmpX, tmpY, tmpZ;
	GLdouble c, s;
#endif

	dy=(x-lastX)*(M_PI/180.0);
	dx=(y-lastY)*(M_PI/180.0);

	lastX=x;
	lastY=y;

#ifndef USE_QUATERNIONS

#ifdef USEABLE_ANGLE
	Xangle+=dx;
	Yangle+=dy;
#else
	newX=Xangle+dy*sin(Zangle*(M_PI/180.0));
	newY=Yangle+dy*cos(Zangle*(M_PI/180.0));
	newZ=Zangle+dx;

	Xangle=newX;
	Yangle=newY;
	Zangle=newZ;
#endif
#else
	/*USE_QUATERNIONS*/
/************************************************************************
rotations:
Qx=[cos(X/2),(sin(X/2),0,0)]
Qy=[cos(Y/2),(0,sin(X/2),0)]
Qz=[cos(Z/2),(0,0,sin(X/2))]

multiplication:
Qw=Q1w*Q2w -Q1x*Q2x -Q1y*Q2y -Q1z*Q2z;
Qx=Q1w*Q2x +Q1x*Q2w +Q1y*Q2z -Q1z*Q2y;
Qy=Q1w*Q2y -Q1x*Q2z +Q1y*Q2w +Q1z*Q2x;
Qz=Q1w*Q2z +Q1x*Q2y -Q1y*Q2x +Q1z*Q2w;
************************************************************************/

	c=cos(dx*0.5);
	s=sin(dx*0.5);

	tmpW=c*Wangle-s*Xangle;
	tmpX=c*Xangle+s*Wangle;
	tmpY=c*Yangle-s*Zangle;
	tmpZ=c*Zangle+s*Yangle;

	c=cos(dy*0.5);
	s=sin(dy*0.5);

	Wangle=c*tmpW-s*tmpY;
	Xangle=c*tmpX+s*tmpZ;
	Yangle=c*tmpY+s*tmpW;
	Zangle=c*tmpZ-s*tmpX;
#endif


	POSTFULLREDISPLAY;
}

void hresize(int h, int w)
{
	height=h;
	width=w;

	glViewport(0, 0, w, h);

	glutPostRedisplay();
}

void regen(void)
{
#ifdef USE_QUATERNIONS
	GLdouble theta, tmp;
#endif
	glPopMatrix();
	glPushMatrix();
	glScalef(4.0, 4.0, 4.0);
#ifndef USE_QUATERNIONS
#ifndef USEABLE_ANGLE
	glRotated(Xangle*(180.0/M_PI), 1.0, 0.0, 0.0);
	glRotated(Yangle*(180.0/M_PI), 0.0, 1.0, 0.0);
	glRotated(Zangle*(180.0/M_PI), 0.0, 0.0, 1.0);
#else
/*	glRotated(Yangle*(180.0/M_PI), 0.0, 1.0, 0.0);*/
	glRotated(Xangle*(180.0/M_PI), 1.0, 0.0, 0.0);
	glRotated(Yangle*(180.0/M_PI), 0.0, 1.0, 0.0);
#endif
#else
/************************************************************************
rotatef:
theta=acos(Q.w)*2;
x=Q.x/sin(theta/2);
y=Q.y/sin(theta/2);
z=Q.z/sin(theta/2);
************************************************************************/

	tmp=acos(Wangle);
	theta=tmp*2.0*180.0/M_PI;
	tmp=1.0/sin(tmp);
	glRotated(theta, Xangle*tmp, Yangle*tmp, Zangle*tmp);
#endif
}

void redisplay(void)
{
	regen();
	glutPostRedisplay();
	glutIdleFunc(NULL);
}

void draw(void)
{
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

	glShadeModel(GL_SMOOTH);
	glPushMatrix();
	glCallList(curstart+curdepth*curskip);
	glPopMatrix();

#ifdef FONTTEST
	glCallList(fore);
#endif

/*	glFlush();*/
	glutSwapBuffers();

}

/* **********************************************************************
*OpenGL Initialization							*
************************************************************************/


int startup(int argc, char **argv)
{
	float tmp[4]={1.0, 1.0, 1.0, 1.0};

	glutInit(&argc, argv);

	glutInitDisplayMode(GLUT_RGBA|GLUT_DOUBLE|GLUT_DEPTH);

	glutInitWindowSize(500, 500);
	glutCreateWindow("Pascal's Pyramid Scheme");

	glutDisplayFunc(draw);
	glutReshapeFunc(hresize);
	glutKeyboardFunc(hkeyb1);
	glutSpecialFunc(hkeyb2);
	glutMouseFunc(hmouseb);
	glutMotionFunc(hmousem);
	glutIdleFunc(redisplay);


#ifdef USE_FOG
	glEnable(GL_FOG);
/*	glFogi(GL_FOG_MODE, GL_LINEAR);*/
	glFogi(GL_FOG_MODE, GL_EXP);
/*	glFogf(GL_FOG_DENSITY, 0.1);
	glFogf(GL_FOG_START, 0.0);
	glFogf(GL_FOG_END, 5.0);*/
#endif
	switch(argc) {
	default:
		fprintf(stderr, "pascal: Excessive arguments\n");
#ifdef USE_FOG
	case 7:
		glFogf(GL_FOG_END, strtof(argv[6], NULL));
	case 6:
		glFogf(GL_FOG_START, strtof(argv[5], NULL));
	case 5:
		glFogf(GL_FOG_DENSITY, strtof(argv[4], NULL));
#endif

	case 4:
		loc[1]=strtod(argv[3], NULL);
	case 3:
		loc[0]=strtod(argv[2], NULL);

	case 2:
		{
			int i;
			double mult;

			mult=strtod(argv[1], NULL);

			for(i=0; i<3; ++i) {
				lightZeroColor[i]*=mult;
				lightOneColor[i]*=mult;
			}
		}
	case 1:
	}


	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_NORMAL_ARRAY);

	glEnable(GL_CULL_FACE);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LIGHTING);
	glMatrixMode(GL_PROJECTION);
	gluPerspective(70.0, 1.0, 1.0, -1.0);
	glMatrixMode(GL_MODELVIEW);
	gluLookAt(0.0, 0.0, 10.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0);
	glPushMatrix();		/*dummy push so we can pop on model recalc*/
	glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, 1);
	glLightfv(GL_LIGHT0, GL_POSITION, lightZeroPosition);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, lightZeroColor);
	glLightf(GL_LIGHT0, GL_CONSTANT_ATTENUATION, 0.1);
	glLightf(GL_LIGHT0, GL_LINEAR_ATTENUATION, 0.05);
	glLightfv(GL_LIGHT1, GL_POSITION, lightOnePosition);
	glLightfv(GL_LIGHT1, GL_DIFFUSE, lightOneColor);
	glEnable(GL_LIGHT0);
	glEnable(GL_LIGHT1);
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, tmp);
	glEnable(GL_RESCALE_NORMAL);
/*	glEnable(GL_NORMALIZE);*/


	lastX=0;
	lastY=0;

	{
		GLint stk, nest;

		glGetIntegerv(GL_MAX_PROJECTION_STACK_DEPTH, &stk);
		glGetIntegerv(GL_MAX_LIST_NESTING, &nest);
		maxdepth=(stk>nest?nest:stk)-1;
		maxdepth=maxdepth<32?maxdepth:32;
	}

	return(0);
}

void showmatr(GLenum m)
{
	double _matr[16]={0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
0.0, 0.0, 0.0, 0.0, 0.0, 0.0}, *matr=_matr;
	unsigned x, y;

	glGetDoublev(m, matr);

	for(y=0; y<=3; ++y) {
		for(x=0; x<=3; ++x) {
			fprintf(stderr, "%f\t", *matr++);
		}
		fprintf(stderr, "\n");
	}
	fprintf(stderr, "\n");
}

#ifdef FONTTEST
void initfont()
{
	GLfloat color[4]={1.0, 1.0, 1.0, 1.0};

	fore=glGenLists(1);

	glNewList(fore, GL_COMPILE);
#ifdef LOCTEST
	glPushMatrix();
#endif

#ifdef LOCTEST
	glLoadIdentity();
#endif

	glColor3fv(color);
	glRasterPos2fv(loc);
	glutBitmapCharacter(FONT, 'Q');

#ifdef LOCTEST
	glPopMatrix();
#endif
	glEndList();
}
#endif

